use cgmath::{vec3, Vector2, Vector3, Zero};

use std::f32::{MAX, MIN};

use crate::aabb::AABB;
use crate::intersection::Intersection;
use crate::ray::Ray;
use crate::triangle::Triangle;

pub struct AccelerationData<'a> {
    recursive: bool,
    triangles: &'a [Triangle],
    data: [AABB; 1024],
}

impl<'a> AccelerationData<'a> {
    pub fn create_tree(input_data: &'a [Triangle], triangles_per_as: u32, recursive: bool) -> Self {
        // we are creating the tree from the bottom up

        // create our slice
        let mut data = [AABB::new(); 1024];

        // first lets create the leaf nodes
        let leaf_nodes = Self::create_leaf_nodes(input_data, triangles_per_as);

        // copy leaf nodes into our final array
        for (index, aabb) in leaf_nodes.iter().enumerate() {
            data[index + 1] = *aabb;
        }

        // create complete tree from leaf nodes
        Self::create_tree_nodes(&mut data, leaf_nodes.len() as u32);

        // return result
        AccelerationData {
            recursive,
            triangles: input_data,
            data,
        }
    }

    pub fn find_closest_hit(&self, ray: &Ray) -> Option<Intersection> {
        let mut closest_index = -1;
        let mut closest_distance = MAX;
        let mut barycentric = Vector2::zero();

        self.traverse_tree(&mut |_, aabb| {
            if aabb.intersect(ray) {
                // check if node is a leaf
                if aabb.start_index != -1 {
                    for i in aabb.start_index..aabb.end_index {
                        let mut t = 0.0;
                        let mut tmp_bary = Vector2::zero();

                        if self.triangles[i as usize].intersect_mt(ray, &mut t, &mut tmp_bary) {
                            if t < closest_distance {
                                closest_index = i as i32;
                                closest_distance = t;
                                barycentric = tmp_bary;
                            }
                        }
                    }
                }

                true
            } else {
                false
            }
        });

        if closest_index == -1 {
            None
        } else {
            Some(Intersection::new(
                closest_distance,
                self.triangles[closest_index as usize].clone(),
                barycentric,
            ))
        }
    }

    pub fn print_tree(&self) {
        self.traverse_tree(&mut |index, aabb| {
            print_aabb(index, aabb);
            true
        })
    }
}

// private
impl<'a> AccelerationData<'a> {
    fn traverse_tree<F>(&self, f: &mut F)
    where
        F: FnMut(usize, &AABB) -> bool,
    {
        if self.recursive {
            self.traverse_tree_recursively(f, 0);
        } else {
            self.traverse_tree_iteratively(f, 0);
        }
    }

    fn traverse_tree_recursively<F>(&self, f: &mut F, index: usize)
    where
        F: FnMut(usize, &AABB) -> bool,
    {
        if (*f)(index, &self.data[index]) {
            let left_child = self.data[index].left_child;
            if left_child != -1 {
                self.traverse_tree_recursively(f, left_child as usize);
            }

            let right_child = self.data[index].right_child;
            if right_child != -1 {
                self.traverse_tree_recursively(f, right_child as usize);
            }
        }
    }

    fn traverse_tree_iteratively<F>(&self, f: &mut F, index: usize)
    where
        F: FnMut(usize, &AABB) -> bool,
    {
        let mut stack = Vec::new();

        // push root node
        stack.push(index);

        while !stack.is_empty() {
            let v = stack.pop().unwrap();

            if (*f)(v, &self.data[v]) {
                let left_child = self.data[v].left_child;
                if left_child != -1 {
                    stack.push(left_child as usize);
                }

                let right_child = self.data[v].right_child;
                if right_child != -1 {
                    stack.push(right_child as usize);
                }
            }
        }
    }

    fn create_leaf_nodes(input_data: &[Triangle], triangles_per_as: u32) -> Vec<AABB> {
        let mut acceleration_data = Vec::new();

        let accel_count = (input_data.len() as f32 / triangles_per_as as f32).ceil() as usize;

        for i in 0..accel_count {
            let mut accel = AABB::new();

            let mut bbmin = vec3(MAX, MAX, MAX);
            let mut bbmax = vec3(MIN, MIN, MIN);

            // 3 vertices per triangles
            // 8 triangles per acceleration structure
            let start_index = triangles_per_as as usize * i;
            let mut end_index = triangles_per_as as usize * (i + 1);

            // check if end_index exceeds input data length
            if end_index > input_data.len() {
                end_index = input_data.len();
            }

            // set index information
            accel.start_index = start_index as isize;
            accel.end_index = end_index as isize;

            // create bounding box
            for k in start_index..end_index {
                for vertex in input_data[k].vertices.iter() {
                    // check minimum values
                    bbmin.x = bbmin.x.min(vertex.position.x);
                    bbmin.y = bbmin.y.min(vertex.position.y);
                    bbmin.z = bbmin.z.min(vertex.position.z);

                    // check maximum values
                    bbmax.x = bbmax.x.max(vertex.position.x);
                    bbmax.y = bbmax.y.max(vertex.position.y);
                    bbmax.z = bbmax.z.max(vertex.position.z);
                }
            }

            accel.bounds[0] = bbmin;
            accel.bounds[1] = bbmax;

            acceleration_data.push(accel);
        }

        acceleration_data
    }

    fn create_tree_nodes(data: &mut [AABB], leaf_node_count: u32) {
        let mut current_level_count = leaf_node_count;
        let mut start_index = 1;

        loop {
            let odd = (current_level_count % 2) == 1;
            let next_level_count = (current_level_count as f32 / 2.0).ceil() as u32;
            let next_level_start = start_index + current_level_count;

            for i in 0..next_level_count {
                let mut aabb = AABB::new();

                let index = (start_index + i * 2) as usize;

                if i == (next_level_count - 1) && odd {
                    aabb.bounds[0] = data[index].bounds[0];
                    aabb.bounds[1] = data[index].bounds[1];
                    aabb.left_child = index as isize;
                } else {
                    aabb.bounds[0] = Self::min(data[index].bounds[0], data[index + 1].bounds[0]);
                    aabb.bounds[1] = Self::max(data[index].bounds[1], data[index + 1].bounds[1]);
                    aabb.left_child = index as isize;
                    aabb.right_child = index as isize + 1;
                }

                if next_level_count == 1 {
                    data[0] = aabb;
                    break;
                } else {
                    data[(next_level_start + i) as usize] = aabb;
                }
            }

            if next_level_count == 1 {
                break;
            }

            current_level_count = next_level_count;
            start_index = next_level_start;
        }
    }

    fn min(v0: Vector3<f32>, v1: Vector3<f32>) -> Vector3<f32> {
        Vector3::new(v0.x.min(v1.x), v0.y.min(v1.y), v0.z.min(v1.z))
    }

    fn max(v0: Vector3<f32>, v1: Vector3<f32>) -> Vector3<f32> {
        Vector3::new(v0.x.max(v1.x), v0.y.max(v1.y), v0.z.max(v1.z))
    }
}

fn print_aabb(index: usize, aabb: &AABB) {
    println!("({}):\t{:?}", index, aabb);
}
